;;;; shift-reset.scm
;;;; Kon Lovett, Apr 6 '06

;; Dynamically scoped shift/reset (Olivier Danvy & Andrzej Filinski)

;FIXME Use per-thread parameters for *meta-k*

(module shift-reset

(;export
  (reset *reset) (shift *shift)
  (reset-values *reset-values) (shift-values *shift-values) )

(import scheme
  (chicken base)
  (chicken type)
  (chicken continuation)
  (only miscmacros let/cc))

(: *shift (procedure -> *))
(: *reset (procedure -> *))
(: *shift-values (procedure -> . *))
(: *reset-values (procedure -> . *))

(define-syntax reset
  (syntax-rules ()
    ((_ BODY ...) (*reset (lambda () BODY ...)) ) ) )

(define-syntax shift
  (syntax-rules ()
    ((_ SP BODY ...) (*shift (lambda (SP) BODY ...)) ) ) )

(define-syntax reset-values
  (syntax-rules ()
    ((_ BODY ...) (*reset-values (lambda () BODY ...)) ) ) )

(define-syntax shift-values
  (syntax-rules ()
    ((_ SP BODY ...) (*shift-values (lambda (SP) BODY ...)) ) ) )

;;
;; Local impl of continuation binding
;;

(define-syntax let/ccp
  (syntax-rules ()
    ((_ K BODY ...) (continuation-capture (lambda (K) BODY ...)) ) ) )

;;
;; Common error
;;

(define (bad-k s r) (error s "no toplevel in scope" r))

;;
;; Single Valued - With Dynamic-Wind
;;

(define *meta-k* (lambda (val) (bad-k 'shift 'reset)))

(define-syntax *return
  (syntax-rules ()
    ((_ EXPR) (*meta-k* EXPR)) ) )

(define (*reset thunk)
  (let ((meta-k *meta-k*))
    (let/cc k
      (set! *meta-k* (lambda (val) (set! *meta-k* meta-k) (k val)))
      (*return (thunk)) ) ) )

(define (*shift proc)
  (let/cc k
    (*return (proc (lambda (val) (*reset (lambda () (k val)))))) ) )

;;
;; Multi-Valued - With Dynamic-Wind
;;

(define *meta-kv* (void))

(define-syntax *return-values
  (syntax-rules ()
    ((_ EXPR CALLER)
      (if (continuation? *meta-kv*) (continuation-graft *meta-kv* (lambda () EXPR))
          (bad-k CALLER 'reset-values) ) ) ) )

(define (*reset-values thunk)
  (let ((meta-kv *meta-kv*))
    (call-with-values
      (lambda ()
        (let/ccp k
          (set! *meta-kv* k)
          (call-with-values
            thunk
            (lambda vals (*return-values (apply values vals) 'reset-values)))))
      (lambda vals
        (set! *meta-kv* meta-kv)
        (apply values vals))) ) )

(define (*shift-values proc)
  (let/ccp k
    (*return-values
      (proc (lambda vals (*reset-values (lambda () (apply continuation-return k vals)))))
      'shift-values) ) )

) ;module shift-reset
